// PID implementation for one motor using serial comm with hyperterm

// PORTA Potentiometer ADC input
// PORTB PWM outputs
// PORTD Direction bits, Hyperterm
// PORTC LEDs position output
 
#include "pid.h"
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <math.h>

//set up the debugging utility ASSERT
#define __ASSERT_USE_STDERR
#include <assert.h>
#include "uart.h"
// UART file descriptor
// putchar and getchar are in uart.c
FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

int init_flag;
volatile char Ain0, Ain1 ; 			//raw A to D numbers
volatile int motorAin0, motorAin1;	//flags for ADCs
volatile int compareTimer;
int16_t motorReference1, motorReference2;
int motor1init, motor2init;

volatile unsigned int pulse_tick_counter, time_counter;
unsigned int pulse_width_0, pulse_width_1, pulse_period;

//the  subroutines
void task1(void);  		// Waits for a string and sets a flag
void task2(void);		// Prints the system time and string from task1 
void getstr_int(void);	// Starts getting a string from serial line
void putstr_int(void);	// Starts a send to serial line
void initialize(void); 	// All the usual mcu stuff 
          
volatile int time1, time2	;	// task scheduling timeout counters
volatile long time;				// system time
unsigned int v;					// number entered by human
char newstr ;					// task1-->task2 handshake
	
// RXC ISR variables								
volatile char r_index;  //current string index
volatile char r_buffer[16];	//input string 
volatile char r_ready;  //flag for receive done
volatile char r_char;  //current character  

//*** WII TEMP VARIABLES
unsigned char r_buffer_x[5], r_buffer_y[5], r_buffer_button[1]; 

// TX ISR variables								
volatile char t_index;  //current string index
volatile char t_buffer[16];	//output string 
volatile char t_ready;  //flag for transmit done
volatile char t_char;  //current character

void init();
int process_x(int);
int process_y(int);

/*! \brief P, I and D parameter values
 *
 * The K_P, K_I and K_D values (P, I and D gains)
 * need to be modified to adapt to the application at hand
 */
//! \xrefitem todo "Todo" "Todo list"
#define K_P     0.4
//! \xrefitem todo "Todo" "Todo list"
#define K_I     0.009
//! \xrefitem todo "Todo" "Todo list"
#define K_D     0.032

//#define K_P		0.25
//#define K_I		0.0
//#define K_D		0.0

/*! \brief Flags for status information
 */
struct GLOBAL_FLAGS {
  //! True when PID control loop should run one time
  uint8_t pidTimer:1;
  uint8_t dummy:7;
} gFlags = {0, 0};

//! Parameters for regulator
struct PID_DATA pidData;

/*! \brief Sampling Time Interval
 *
 * Specify the desired PID sample time interval
 * With a 8-bit counter (255 cylces to overflow), the time interval value is calculated as follows:
 * TIME_INTERVAL = ( desired interval [sec] ) * ( frequency [Hz] ) / 255
 */
//! \xrefitem todo "Todo" "Todo list"
#define TIME_INTERVAL   1000

//timeout values for each task
#define t1 10000  
#define t2 20000  

//I like these definitions
#define begin {
#define end   }

//**********************************************************
// ADC ISR to toggle different input pins
ISR(ADC_vect)
{
	if (motorAin0) {
		
		//enable adc1 for next sample
		ADMUX = 0b00100001;
		//get the last sample
		Ain0 = ADCH;
		//PORTC=~((int)Ain0);
		//move on to next read
		motorAin0=0;
		motorAin1=1;
	}
	else if (motorAin1) {
		//enable adc0 for next sample
		ADMUX = 0b00100000;
		//get the last sample
		Ain1 = ADCH;
		//move on to next read
		//PORTC=~((int)Ain1);
		motorAin0=1;
		motorAin1=0;
	}

 	//start another conversion
 	ADCSRA |= (1<<ADSC) ; 
	
}

// Timer0 interrupt, increase 1 every 100us
ISR (TIMER0_COMPA_vect) 
{
	if (compareTimer > 0) compareTimer--;
	if (time1>0)	--time1;
	if (time2>0) 	--time2;
	
	//PWM counters
	pulse_tick_counter++;
	time_counter++;
	
}    

//**********************************************************
// UART character-ready ISR
// builds a sting and signals when the string is complete
// supports backspace

ISR (USART0_RX_vect)
{
	r_char = UDR0 ;    //get a char
	UDR0 = r_char;    //then print it
	//build the input string
	if (r_char != '\r') 	// Is the input a <enter>?
	{ 	
		//*** HYPERTERM INPUT
		if (r_char == '\b')	// Is the input a backspace? 
		{
			putchar(' '); 	// erase the character on the screen
			putchar('\b');	// backup
			--r_index ;		// wipe a character from the string
		}
		else   
 			r_buffer[r_index++] = r_char ; // add a character to the string
    }
    else  //end of line
    {              
        putchar('\n');  			//use putchar to avoid overwrite
		r_buffer[r_index] = 0x00;   //zero terminate
		r_ready = 1;				//signal cmd processor
		UCSR0B ^= (1<<RXCIE0) ;   	//stop rec ISR -- clear rxc
	}
}

/**********************************************************/
//UART xmit-empty ISR
ISR (USART0_UDRE_vect)
begin
	t_char = t_buffer[++t_index];
	if (t_char == 0) 			// end of string?	
	begin
		UCSR0B ^= (1<<UDRIE0) ; 	// kill isr -- clear tx enable
		t_ready = 1; 			// transmit done 
	end
	else	UDR0 = t_char ;  	//send the char 
end
  
//**********************************************************          
//Task 1   input a string and print it
void task1(void) 
begin 
  //print ad get another serial string      
  if (r_ready)    
  begin
  	// read string which is now ready
	sscanf(r_buffer,"%d",&v);
	// set up procedure to get the next string input
	// and read it using receive ISR
 	getstr_int();
	// signal task2 that a new string is ready
	newstr = 1;     
  end     
end  
 
//********************************************************** 
//Task 2  print the meaasge and system time in seconds
void task2(void) 
begin
  // see if there is a message from task1 and if the last transmit is done
  if (t_ready && newstr)
     begin
	   // make a string of system time (sec) and number from task 1
       sprintf(t_buffer,"%ld %d\n\r",time/1000, v) ; 
	   // print it using transmit ISR 
       putstr_int();
	   // clear the string ready handshake from task 1 
       newstr = 0;
     end
end  

// do this??
//**********************************************************
// Prints the string strLog in hyperterminal, waiting for t_ready to be 1.
void printLine(unsigned char *strLog) {
	// infinite loop guarentees that we wait for previous transmissions to 
	// complete before executing this
	while(!t_ready) {}
	strcpy(t_buffer, strLog);
	putstr_int();
}
  
//********************************************************** 
// Non-blocking keyboard input: initializes ISR-driven receive.
// This routine merely sets up the ISR, which then
//does all the work of getting a command.
void getstr_int(void) 
begin
  r_ready=0;	// mark string as not ready
  r_index=0;	// reset index
  // turn on receive ISR
  UCSR0B |= (1<<RXCIE0) ;
end

//********************************************************** 
// Nonblocking print: initializes ISR-driven transmit. 
// This routine merely sets up the ISR, then
// sends one character, The ISR does all the work.   
void putstr_int(void) 
begin
  t_ready=0;	// mark transmitter as busy
  t_index=0;	// reset index
  // see if there is actually a string
  if (t_buffer[0]>0) 
  begin
  	// if so, send one chararcter
  	putchar(t_buffer[0]);
	// and turn on transmit (UDR empty) ISR
  	UCSR0B |= (1<<UDRIE0) ;
 end 
end

// Read reference/desired value from keyboard
void Get_Reference(void)
{	
	// Parse desired motor positions
	if(r_ready==1) {
		char mref[3];
		char x_coord[4];
		char y_coord[3];
		int i=1;
		int j=0;
		int k=0;
		//**GUI INPUT
		//If desired value for motor 1 is entered
		if (r_buffer[0] == 'x') {
			while (i < r_index) {
				mref[i-1] = r_buffer[i];
				i++;
			}
			motor1init = 1;
			motorReference1 = atoi(mref);
		}
		// desired value for motor 2`
		else if (r_buffer[0] == 'y') {
			while (i < r_index) {
				mref[i-1] = r_buffer[i];
				i++;
			}
			motor2init = 1;
			motorReference2 = atoi(mref);
		}

		//**WII INPUT
		else {
			// light up an LED for debugging
			PORTD |= (1<<6);
			while (r_buffer[j] != ' ') {
				x_coord[j] = r_buffer[j];
				j++;
			}
			while (j < r_index) {
				y_coord[k] = r_buffer[j+1];
				j++;
				k++;
			}
			
			motor1init=1;
			motorReference1 = process_x(atoi(x_coord));
			PORTC = ~motorReference1;
			motor2init=1;
			motorReference2 = process_y(atoi(y_coord));
			PORTC = ~motorReference2;
		}
		
		r_ready=0;
		getstr_int();
	}	
}

// Process Wii IR coordinates
int process_x(int x_coord) {
	// x IR coordinates range from 0-1023, mapping to
	// potentiometer values 0-255 from ADC.
	// A reading of 0 means no signal present, so
	// just keep where it is
	if (x_coord == 0)	return motorReference1;
	else {
		if (x_coord < 100)		return 15;	
		else if (x_coord < 250)	return 32;
		else if (x_coord < 350)	return 64;
		else if (x_coord < 450)	return 96;
		else if (x_coord < 550)	return 128;
		else if (x_coord < 650)	return 160;
		else if (x_coord < 750)	return 192;
		else if (x_coord < 900)	return 224;
		else 					return 240;
	}
}

int process_y(int y_coord) {
	// y IR coordinates range from 0-767, mapping to
	// potentiometer values 0-255 from ADC
	if (y_coord == 0)	return motorReference2;
	else {
		if (y_coord < 80)		return 15;	
		else if (y_coord < 184)	return 32;
		else if (y_coord < 240)	return 64;
		else if (y_coord < 320)	return 96;
		else if (y_coord < 400)	return 128;
		else if (y_coord < 480)	return 160;
		else if (y_coord < 560)	return 192;
		else if (y_coord < 664)	return 224;
		else 					return 240;
	}
}

// Set PWM duty cycle depending on distance from desired position
void Set_Input0(int16_t inputValue)
{
	pulse_width_0 = inputValue;
}

void Set_Input1(int16_t inputValue)
{
	pulse_width_1 = inputValue;
}

int main(void)
{
	int16_t measurementValue, inputValue;
	init();
  
  	//set pulse widths for different speeds
	pulse_period = 25;
	pulse_width_0 = 0;
	pulse_width_1 = 0;
	pulse_tick_counter = 0;

  while(1){
  
	//** KEY
	// Direction pins: d.2 and d.3
	// Brake pins: d.4 and d.5
	// PWM pins: b.0 and b.1
  
    if (time1==0){time1=t1;	task1(); }
    //if (time2==0){time2=t2;	task2(); }
 
	// Run PID calculations once every PID timer timeout
    if(compareTimer==0)
    {
		//Acquire desired position reference values from hyperterm
		Get_Reference();

		//*** MOTOR 0 ***
		measurementValue = (int16_t)Ain0;
		
		if (motor1init == 0) motorReference1 = measurementValue;
		  
		// Invert direction (counter-clockwise)
		if (motorReference1 < measurementValue) {
			//set d.2 to 1
			PORTD |= (1<<2);
		}
		else {
			//set d.2 to 0
			PORTD &= ~(1<<2);
		}

		inputValue = pid_Controller(motorReference1, measurementValue, &pidData);
			
		// If error is small enough, brake
		// port d.4
		if (abs(motorReference1-measurementValue) < 10) PORTD |= (1<<4);
		else PORTD &= ~(1<<4);
			
		Set_Input0(inputValue);

		//*** MOTOR 1 ***
		measurementValue = (int16_t)Ain1;
		
		if (motor2init == 0) motorReference2 = measurementValue;
		  
		// Invert direction (counter-clockwise)
		if (motorReference2 < measurementValue) {
			//set d.3 to 1
			PORTD |= (1<<3);
		}
		else {
			PORTD &= ~(1<<3);
		}

		inputValue = pid_Controller(motorReference2, measurementValue, &pidData);
			
		// If error is small enough, brake
		// port d.5
		if (abs(motorReference2-measurementValue) < 10) PORTD |= (1<<5);
		else PORTD &= ~(1<<5);
			
		Set_Input1(inputValue);
		
		// Restart
		compareTimer=TIME_INTERVAL;
    }
	
	// Power motors depending on speed/pulse_width
	if (pulse_tick_counter < pulse_period) {
		if (pulse_tick_counter >= pulse_width_0) {
			PORTB &= ~(1<<0);
		}
		if (pulse_tick_counter >= pulse_width_1) {
			PORTB &= ~(1<<1);
		}
	}
	else {
		// pull up ALL IO pin
		pulse_tick_counter = 0;
		PORTB=0xff;
	}
  }
}

// Setting it all up
void init(void)
{

	// init the UART -- uart_init() is in uart.c
	uart_init();
	stdout = stdin = stderr = &uart_str;
	fprintf(stdout,"Starting ISR UART demo...\n\r"); 

	pid_Init(K_P * SCALING_FACTOR, K_I * SCALING_FACTOR , K_D * SCALING_FACTOR , &pidData);

	//set up timer 0 for 100 uSec timebase 
	TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR 
	OCR0A = 200;  		//set the compare reset to 200 ticks (100us)
	//set prescalar to divide by 8, 1/2 us ticks
	TCCR0B= 2; 
	// turn on clear-on-match
	TCCR0A= (1<<WGM01) ;

	//init the A to D converter 
	//channel zero/ left adj /EXTERNAL Aref
	//!!!CONNECT Aref jumper!!!!
	ADMUX = (1<<ADLAR) ;  
	
	//enable ADC and set prescaler to 1/128*16MHz=125,000
	//and set interupt enable
	//and start a conversion
	ADCSRA = (1<<ADEN) | (1<<ADSC) | (1 << ADIE) + 7 ; 
	   
	//start by reading motor0's input
	motorAin0=1;
	motorAin1=0;
   
	// init the task timers
	time1=t1;
	time2=t2;    
	compareTimer=TIME_INTERVAL;
	
	// initialize the USRT handshake flags
	r_ready=0; // initially, there is no input ready
	t_ready=1; // initially, the transmitter is ready
	  
	//Set up port B, output to motors
	DDRB=0xFF;
	PORTB=0;
	
	//Set up port D, output direction bit
	DDRD=0xff;
	PORTD=0;
	
	//port C, output to LEDs
	DDRC=0xff;
	PORTC=0xff;
	
	sei();
	
	// get the first string from the human
	getstr_int();
}
